
#include "x.h"
#include "xt.h"
#include "xpex.h"
#include <math.h>
#include <stdlib.h>
#include "book_utils.h"
#include "externs.h"
#include "defs.h"
#include "globals.h"

Display            *dpy;
Window            window;
PEXRenderer        renderer;
PEXRendererAttributes    rattrs;
PEXStructure    top_struct, sphere;
Atom        delete_window_atom;
Widget    toplevel, canvas;

extern Pixmap mark;

typedef struct _Text {
    char *direction;
    char *point;
} Text, *TextPtr;

#define NORMALIZE( _v )                          \
{                                                \
    double mag;                                  \
                                                 \
    mag = 1.0 / sqrt((_v).x * (_v).x             \
          + (_v).y * (_v).y + (_v).z * (_v).z);  \
    (_v).x *= mag, (_v).y *= mag, (_v).z *= mag; \
}


/* Callback function to quit the program. */
void
quit( w, client_data, call_data )
Widget     w;
caddr_t    client_data;
XtPointer *call_data;
{
#if 0
    XClientMessageEvent        *cm = &call_data->event->xclient;

    if ( cm->data.l[0] == delete_window_atom ) {
        XtDestroyWidget( toplevel );
        exit( 0 );
    }
#endif
}


/* Callback for Expose events. */
void
expose( w, client_data, event )
Widget    w;
XtPointer    client_data;
XEvent    *event;
{
    /* Redraw when last Expose event received. */
    if ( event->xexpose.count == 0 )
        redraw();
}



void
set_lights( widget, choice )
Widget    widget;
int        choice;
{
    extern int light_switches[4];
    extern PEXLightEntry lights[4];
    extern int cur_light;

    Boolean vector_on = light_switches[1],
            point_on = light_switches[2],
            spot_on = light_switches[3];

    int           i, on_off;
    PEXCoord      pln[2];
    Widget        button;
    extern Widget lightEntry[8];
    extern int num_on, num_off;
    extern PEXTableIndex light_on[4];
    extern PEXTableIndex light_off[4];

    if ( choice == 0 || choice == 7 ) {
        if ( choice == 0 ) { /* Set all lights. */
            on_off = 1;
            for( i= 2; i< 6; i++ )
                XtVaSetValues( lightEntry[i], XtNleftBitmap, mark, NULL );
            PEXSetElementPtr( dpy, top_struct, PEXBeginning, 0 );
            PEXDeleteBetweenLabels( dpy, top_struct,
                START_LIGHT_LABEL, END_LIGHT_LABEL );
            PEXSetEditingMode( dpy, top_struct, PEXStructureInsert );

            PEXLabel( dpy, top_struct, PEXOCStore, DIR_LIGHT_LABEL );
            pln[0].x = (lights[DIRECTIONAL].direction.x<0.) ? -lights[DIRECTIONAL].direction.x+1. : -lights[DIRECTIONAL].direction.x-1.;
            pln[0].y = (lights[DIRECTIONAL].direction.y<0.) ? -lights[DIRECTIONAL].direction.y+1. : -lights[DIRECTIONAL].direction.y-1.;
            pln[0].z = (lights[DIRECTIONAL].direction.z<0.) ? -lights[DIRECTIONAL].direction.z+1. : -lights[DIRECTIONAL].direction.z-1.;
            pln[0].x = (lights[DIRECTIONAL].direction.x==0.) ? 0. : pln[0].x;
            pln[0].y = (lights[DIRECTIONAL].direction.y==0.) ? 0. : pln[0].y;
            pln[0].z = (lights[DIRECTIONAL].direction.z==0.) ? 0. : pln[0].z;
            pln[1].x = -lights[DIRECTIONAL].direction.x;
            pln[1].y = -lights[DIRECTIONAL].direction.y;
            pln[1].z = -lights[DIRECTIONAL].direction.z;
            PEXPolyline( dpy, top_struct, PEXOCStore, 2, pln );

            PEXLabel( dpy, top_struct, PEXOCStore, POS_LIGHT_LABEL );
            PEXMarkers( dpy, top_struct, PEXOCStore,
                1, &lights[POSITIONAL].point );

            PEXLabel( dpy, top_struct, PEXOCStore, SPT_LIGHT_LABEL );
            pln[0].x = lights[SPOT].point.x;
            pln[0].y = lights[SPOT].point.y;
            pln[0].z = lights[SPOT].point.z;
            pln[1].x = lights[SPOT].point.x + lights[SPOT].direction.x;
            pln[1].y = lights[SPOT].point.y + lights[SPOT].direction.y;
            pln[1].z = lights[SPOT].point.z + lights[SPOT].direction.z;
            PEXPolyline( dpy, top_struct, PEXOCStore, 2, pln );

            vector_on = point_on = spot_on = TRUE;
        }
        else {
            on_off = 0;
            for( i= 2; i< 6; i++ )
                XtVaSetValues( lightEntry[i], XtNleftBitmap, None, NULL );
            PEXSetEditingMode( dpy, top_struct, PEXStructureReplace );
            PEXSetElementPtr( dpy, top_struct, PEXBeginning, 0 );
            PEXSetElementPtrAtLabel( dpy, top_struct, DIR_LIGHT_LABEL, 1 );
            PEXNoop( dpy, top_struct, PEXOCStore );
            PEXSetElementPtrAtLabel( dpy, top_struct, POS_LIGHT_LABEL, 1 );
            PEXNoop( dpy, top_struct, PEXOCStore );
            PEXSetElementPtrAtLabel( dpy, top_struct, SPT_LIGHT_LABEL, 1 );
            PEXNoop( dpy, top_struct, PEXOCStore );

            vector_on = point_on = spot_on = FALSE;
        }
        for ( i = 0; i < 4 ; i++ )
            light_switches[i] = on_off;
    }
    else {
        light_switches[choice - 2] = !light_switches[choice - 2];
        if( light_switches[choice - 2] )
            XtVaSetValues( widget, XtNleftBitmap, mark, NULL );
        else
            XtVaSetValues( widget, XtNleftBitmap, None, NULL );

        PEXSetEditingMode( dpy, top_struct, PEXStructureReplace );
        PEXSetElementPtr( dpy, top_struct, PEXBeginning, 0 );
        switch( choice ) {
        case 3:    /* vector */
            PEXSetElementPtrAtLabel( dpy, top_struct, DIR_LIGHT_LABEL, 1 );
            if( vector_on )
                PEXPolyline( dpy, top_struct, PEXOCStore, 0, (PEXCoord *)0 );
            else {
                pln[0].x = (lights[DIRECTIONAL].direction.x<0.) ? -lights[DIRECTIONAL].direction.x+1. : -lights[DIRECTIONAL].direction.x-1.;
                pln[0].y = (lights[DIRECTIONAL].direction.y<0.) ? -lights[DIRECTIONAL].direction.y+1. : -lights[DIRECTIONAL].direction.y-1.;
                pln[0].z = (lights[DIRECTIONAL].direction.z<0.) ? -lights[DIRECTIONAL].direction.z+1. : -lights[DIRECTIONAL].direction.z-1.;
                pln[0].x = (lights[DIRECTIONAL].direction.x==0.) ? 0.: pln[0].x;
                pln[0].y = (lights[DIRECTIONAL].direction.y==0.) ? 0.: pln[0].y;
                pln[0].z = (lights[DIRECTIONAL].direction.z==0.) ? 0.: pln[0].z;
                pln[1].x = -lights[DIRECTIONAL].direction.x;
                pln[1].y = -lights[DIRECTIONAL].direction.y;
                pln[1].z = -lights[DIRECTIONAL].direction.z;
                PEXPolyline( dpy, top_struct, PEXOCStore, 2, pln );
            }
            vector_on = !vector_on;
            break;
        case 4:    /* point */
            PEXSetElementPtrAtLabel( dpy, top_struct, POS_LIGHT_LABEL, 1 );
            if( point_on )
                PEXMarkers( dpy, top_struct, PEXOCStore, 0, NULL );
            else
                PEXMarkers( dpy, top_struct, PEXOCStore, 1, &lights[POSITIONAL].point );
            point_on = !point_on;
            break;
        case 5:    /* spot */
            PEXSetElementPtrAtLabel( dpy, top_struct, SPT_LIGHT_LABEL, 1 );
            if( spot_on )
                PEXPolyline( dpy, top_struct, PEXOCStore, 0, (PEXCoord *)0 );
            else {
                pln[0].x = lights[SPOT].point.x;
                pln[0].y = lights[SPOT].point.y;
                pln[0].z = lights[SPOT].point.z;
                pln[1].x = lights[SPOT].point.x + lights[SPOT].direction.x;
                pln[1].y = lights[SPOT].point.y + lights[SPOT].direction.y;
                pln[1].z = lights[SPOT].point.z + lights[SPOT].direction.z;
                PEXPolyline( dpy, top_struct, PEXOCStore, 2, pln );
            }
            spot_on = !spot_on;
            break;
        default:
            break;
        }
    }

    num_on = num_off = 0;
    for ( i = 0; i < 4; i++ )
        if ( light_switches[i] )
            light_on[num_on++] = i + 1;
        else
            light_off[num_off++] = i + 1;

    PEXSetEditingMode( dpy, top_struct, PEXStructureReplace );
    PEXSetElementPtr( dpy, top_struct, PEXBeginning, 0 );
    PEXSetElementPtrAtLabel( dpy, top_struct, LIGHT_STATE_LABEL, 1 );
    PEXSetLightSourceState( dpy, top_struct, PEXOCStore,
        num_on, light_on, num_off, light_off );
    redraw();
}


void
set_shading( widget, choice )
Widget    widget;
int        choice;
{
    int shading_method;
    extern int cur_shading;
    int i;
    extern Widget shadeEntry[4];

    switch ( choice ) {
    case 0: shading_method = PEXSurfaceInterpNone;			break;
    case 1: shading_method = PEXSurfaceInterpColor;			break;
    case 2: shading_method = PEXSurfaceInterpDotProduct;	break;
    case 3: shading_method = PEXSurfaceInterpNormal;		break;
    }
    for( i= 0; i< 4; i++ )
        XtVaSetValues( shadeEntry[i], XtNleftBitmap, None, NULL );
    XtVaSetValues( widget, XtNleftBitmap, mark, NULL );

    if ( shading_method != cur_shading ) {
        cur_shading = shading_method;
        PEXSetEditingMode( dpy, top_struct, PEXStructureReplace );
        PEXSetElementPtr( dpy, top_struct, PEXBeginning, 0 );
        PEXSetElementPtrAtLabel( dpy, top_struct, SHADING_METHOD_LABEL, 1 );
        PEXSetSurfaceInterpMethod( dpy, top_struct, PEXOCStore,
            cur_shading );
        PEXSetElementPtr( dpy, top_struct, PEXCurrent, 1 );
        PEXSetBFSurfaceInterpMethod( dpy, top_struct, PEXOCStore,
            cur_shading );
        redraw();
    }
}


void
set_reflectance( widget, choice )
Widget    widget;
int        choice;
{
    int refl;
    int i;
    extern Widget reflectEntry[4];
    extern int cur_refl;

    switch ( choice ) {
    case 0: refl = PEXReflectionNone;		break;
    case 1: refl = PEXReflectionAmbient;	break;
    case 2: refl = PEXReflectionDiffuse;	break;
    case 3: refl = PEXReflectionSpecular;	break;
    }
    for( i= 0; i< 4; i++ )
        XtVaSetValues( reflectEntry[i], XtNleftBitmap, None, NULL );
    XtVaSetValues( widget, XtNleftBitmap, mark, NULL );

    if ( refl != cur_refl ) {
        cur_refl = refl;
        PEXSetEditingMode( dpy, top_struct, PEXStructureReplace );
        PEXSetElementPtr( dpy, top_struct, PEXBeginning, 0 );
        PEXSetElementPtrAtLabel( dpy, top_struct, REFL_EQN_LABEL, 1 );
        PEXSetReflectionModel( dpy, top_struct, PEXOCStore, cur_refl );
        PEXSetElementPtr( dpy, top_struct, PEXCurrent, 1 );
        PEXSetBFReflectionModel( dpy, top_struct, PEXOCStore, cur_refl );
        redraw();
    }
}


void
set_depth_cue( widget, choice )
Widget    widget;
int        choice;
{
    PEXSwitch mode;
    extern PEXDepthCueEntry depth_cue_rep;

    mode = choice == 1 ? PEXOn : PEXOff;
    if ( mode != depth_cue_rep.mode ) {
        depth_cue_rep.mode = mode;
        PEXSetTableEntries( dpy, rattrs.depth_cue_table, 1, 1,
            PEXLUTDepthCue, &depth_cue_rep );
        redraw();
    }
}


void
set_hlhsr( widget, choice )
Widget    widget;
int        choice;
{
           int mode;
    extern int cur_hlhsr_mode;

    mode = choice == 1 ? PEXHLHSRZBuffer : PEXHLHSROff;
    if ( mode != cur_hlhsr_mode ) {
        cur_hlhsr_mode = mode;
        rattrs.hlhsr_mode = mode;
        PEXChangeRenderer( dpy, renderer, PEXRAHLHSRMode, &rattrs );
        redraw();
    }
}


void set_light_color( cbs )
char *cbs;
{
    PEXColorSpecifier *colorSpec;
    extern Widget lightColorLabel;
    String string = (String) malloc(40);
    char type[3];
    float val1, val2, val3;

    switch ( cur_light ) {
    case AMBIENT:
        colorSpec = &lights[AMBIENT].color;
        break;
    case DIRECTIONAL:
        colorSpec = &lights[DIRECTIONAL].color;
        break;
    case POSITIONAL:
        colorSpec = &lights[POSITIONAL].color;
        break;
    case SPOT:
        colorSpec = &lights[SPOT].color;
        break;
    }

    sscanf( cbs, "%s %f %f %f\n", &type, &val1, &val2, &val3 );
    if( !strcmp( type, "RGB" ) ) {
        colorSpec->type = PEXColorTypeRGB;
        colorSpec->value.rgb.red = val1;
        colorSpec->value.rgb.green = val2;
        colorSpec->value.rgb.blue = val3;
        sprintf( string, "R: %.2f G: %.2f B: %.2f", val1, val2, val3 );
        XtVaSetValues( lightColorLabel, XtNlabel, string, NULL );
        free( string );
    }
    else if( !strcmp( type, "HSV" ) ) {
        colorSpec->type = PEXColorTypeHSV;
        colorSpec->value.hsv.hue = val1;
        colorSpec->value.hsv.saturation = val2;
        colorSpec->value.hsv.value = val3;
        sprintf( string, "H: %.2f S: %.2f V: %.2f", val1, val2, val3 );
        XtVaSetValues( lightColorLabel, XtNlabel, string, NULL );
        free( string );
    }
    else if( !strcmp( type, "HLS" ) ) {
        colorSpec->type = PEXColorTypeHLS;
        colorSpec->value.hls.hue = val1;
        colorSpec->value.hls.lightness = val2;
        colorSpec->value.hls.saturation = val3;
        sprintf( string, "H: %.2f L: %.2f S: %.2f", val1, val2, val3 );
        XtVaSetValues( lightColorLabel, XtNlabel, string, NULL );
        free( string );
    }
    else if( !strcmp( type, "CIE" ) ) {
        colorSpec->type = PEXColorTypeCIE;
        colorSpec->value.cie.x = val1;
        colorSpec->value.cie.y = val2;
        colorSpec->value.cie.z = val3;
        sprintf( string, "X: %.2f Y: %.2f Z: %.2f", val1, val2, val3 );
        XtVaSetValues( lightColorLabel, XtNlabel, string, NULL );
        free( string );
    }
    PEXSetTableEntries( dpy, rattrs.light_table, cur_light + 1, 1,
        PEXLUTLight, &lights[cur_light] );
    redraw();
}


void
set_light_source( widget, parameter, cbs )
Widget     widget;
caddr_t    parameter;
XtPointer *cbs;
{
    PEXColor *color;
    float    *atten1, *atten2;
    String string = (String) malloc(40);
    int   set_light = 1;
    float per = *( float* )cbs;
    float top = ( float )per;
 
    extern Widget lightSpreadAngleLabel, lightConcentrationLabel;
                  lightAttenuation1Scroll, lightAttenuation2Scroll;

    switch ( cur_light ) {
    case AMBIENT:
        color =  &lights[AMBIENT].color.value;
        break;
    case DIRECTIONAL:
        color =  &lights[DIRECTIONAL].color.value;
        break;
    case POSITIONAL:
        color =  &lights[POSITIONAL].color.value;
        atten1 = &lights[POSITIONAL].attenuation1;
        atten2 = &lights[POSITIONAL].attenuation2;
        break;
    case SPOT:
        color =  &lights[SPOT].color.value;
        atten1 = &lights[SPOT].attenuation1;
        atten2 = &lights[SPOT].attenuation2;
        break;
    }

    switch ( (int)parameter ) {
    case ATTENUATION_C1:
        if ( cur_light == POSITIONAL || cur_light == SPOT ) {
            *atten1 = top;
            sprintf( string, "%.2f", top );
            XtVaSetValues( lightAttenuation1Label, XtNlabel, string, NULL );
            free( string );
        }
        else
            set_light = 0;
        break;
    case ATTENUATION_C2:
        if ( cur_light == POSITIONAL || cur_light == SPOT ) {
            *atten2 = top;
            sprintf( string, "%.2f", top );
            XtVaSetValues( lightAttenuation2Label, XtNlabel, string, NULL );
            free( string );
        }
        else
            set_light = 0;
        break;
    case SPREAD_ANGLE:
        if ( cur_light == SPOT ) {
            lights[SPOT].spread_angle = top * M_PI/2;
            sprintf( string, "%.0f", 90 * top );
            XtVaSetValues( lightSpreadAngleLabel, XtNlabel, string, NULL );
            free( string );
        }
        else
            set_light = 0;
        break;
    case CONCENTRATION:
        if ( cur_light == SPOT ) {
            lights[SPOT].concentration = 200 * top;
            sprintf( string, "%.0f", 200 * top );
            XtVaSetValues( lightConcentrationLabel, XtNlabel, string, NULL );
            free( string );
        }
        else
            set_light = 0;
        break;
    }

    if ( set_light ) {
        PEXSetTableEntries( dpy, rattrs.light_table, cur_light + 1, 1,
            PEXLUTLight, &lights[cur_light] );
        redraw();
    }
}


void
set_dir_pnt( widget, parameter, cbs )
Widget     widget;
caddr_t    parameter;
XtPointer *cbs;
{
    extern TextPtr text;
    String buffer = (String) malloc(40);

    switch ( (int)parameter ) {
    case DIRECTION:
        if ( cur_light == DIRECTIONAL ) {
            PEXCoord pln[2];

            XtVaGetValues( lightDirectionText, XtNstring, &buffer, NULL );
            sscanf( buffer, "x: %f y: %f z: %f", &lights[cur_light].direction.x,
               &lights[cur_light].direction.y, &lights[cur_light].direction.z );
            NORMALIZE(lights[cur_light].direction);

            PEXSetEditingMode( dpy, top_struct, PEXStructureReplace );
            PEXSetElementPtr( dpy, top_struct, PEXBeginning, 0 );
            PEXSetElementPtrAtLabel( dpy, top_struct, DIR_LIGHT_LABEL, 1 );
            pln[0].x = (lights[cur_light].direction.x<0.) ? -lights[cur_light].direction.x+1. : -lights[cur_light].direction.x-1.;
            pln[0].y = (lights[cur_light].direction.y<0.) ? -lights[cur_light].direction.y+1. : -lights[cur_light].direction.y-1.;
            pln[0].z = (lights[cur_light].direction.z<0.) ? -lights[cur_light].direction.z+1. : -lights[cur_light].direction.z-1.;
            pln[0].x = (lights[cur_light].direction.x==0.) ? 0. : pln[0].x;
            pln[0].y = (lights[cur_light].direction.y==0.) ? 0. : pln[0].y;
            pln[0].z = (lights[cur_light].direction.z==0.) ? 0. : pln[0].z;
            pln[1].x = -lights[cur_light].direction.x;
            pln[1].y = -lights[cur_light].direction.y;
            pln[1].z = -lights[cur_light].direction.z;
            PEXPolyline( dpy, top_struct, PEXOCStore, 2, pln );
        }
        if ( cur_light == SPOT ) {
            PEXCoord pln[2];

            XtVaGetValues( lightDirectionText, XtNstring, &buffer, NULL );
            sscanf( buffer, "x: %f y: %f z: %f", &lights[cur_light].direction.x,
               &lights[cur_light].direction.y, &lights[cur_light].direction.z );
            NORMALIZE(lights[cur_light].direction);

            PEXSetEditingMode( dpy, top_struct, PEXStructureReplace );
            PEXSetElementPtr( dpy, top_struct, PEXBeginning, 0 );
            PEXSetElementPtrAtLabel( dpy, top_struct, SPT_LIGHT_LABEL, 1 );
            pln[0].x = lights[cur_light].point.x;
            pln[0].y = lights[cur_light].point.y;
            pln[0].z = lights[cur_light].point.z;
            pln[1].x = lights[cur_light].point.x+lights[cur_light].direction.x;
            pln[1].y = lights[cur_light].point.y+lights[cur_light].direction.y;
            pln[1].z = lights[cur_light].point.z+lights[cur_light].direction.z;
            PEXPolyline( dpy, top_struct, PEXOCStore, 2, pln );
        }
        break;
    case POINT:
        if ( cur_light == POSITIONAL ) {
            XtVaGetValues( lightPointText, XtNstring, &buffer, NULL );
            sscanf( buffer, "x: %f y: %f z: %f", &lights[cur_light].point.x,
               &lights[cur_light].point.y, &lights[cur_light].point.z );

            PEXSetEditingMode( dpy, top_struct, PEXStructureReplace );
            PEXSetElementPtr( dpy, top_struct, PEXBeginning, 0 );
            PEXSetElementPtrAtLabel( dpy, top_struct, POS_LIGHT_LABEL, 1 );
            PEXMarkers( dpy, top_struct, PEXOCStore, 1, &lights[cur_light].point );
        }
        if ( cur_light == SPOT ) {
            PEXCoord pln[2];

            XtVaGetValues( lightPointText, XtNstring, &buffer, NULL );
            sscanf( buffer, "x: %f y: %f z: %f", &lights[cur_light].point.x,
               &lights[cur_light].point.y, &lights[cur_light].point.z );

            PEXSetEditingMode( dpy, top_struct, PEXStructureReplace );
            PEXSetElementPtr( dpy, top_struct, PEXBeginning, 0 );
            PEXSetElementPtrAtLabel( dpy, top_struct, SPT_LIGHT_LABEL, 1 );
            pln[0].x = lights[cur_light].point.x;
            pln[0].y = lights[cur_light].point.y;
            pln[0].z = lights[cur_light].point.z;
            pln[1].x = lights[cur_light].point.x+lights[cur_light].direction.x;
            pln[1].y = lights[cur_light].point.y+lights[cur_light].direction.y;
            pln[1].z = lights[cur_light].point.z+lights[cur_light].direction.z;
            PEXPolyline( dpy, top_struct, PEXOCStore, 2, pln );
        }
        break;
    }
    PEXSetTableEntries( dpy, rattrs.light_table, cur_light + 1, 1,
        PEXLUTLight, &lights[cur_light] );
    redraw();
}


void
select_cur_light( widget, choice )
Widget    widget;
int        choice;
{
    int i;
    String string = (String) malloc(40);
    extern Widget lightTypeEntry[5];
    extern Widget lightColorLabel, lightDirectionText, lightPointText;
    extern Widget                  lightDirectionTextOk, lightPointTextOk;
    extern Widget lightAttenuation1Scroll, lightAttenuation2Scroll;
    extern Widget lightAttenuation1Label, lightAttenuation2Label;
    
    for( i=0; i< 4; i++ )
        XtVaSetValues( lightTypeEntry[i], XtNleftBitmap, None, NULL );
    XtVaSetValues( widget, XtNleftBitmap, mark, NULL );

    switch ( choice ) {
    case 0:
        cur_light = AMBIENT;
        XtVaSetValues( XtParent(XtParent(widget)), XtNlabel, "Ambient", NULL );
        XtUnmanageChild( lightDirection );
        XtUnmanageChild( lightDirectionText );
        XtUnmanageChild( lightDirectionTextOk );
        XtUnmanageChild( lightPoint );
        XtUnmanageChild( lightPointText );
        XtUnmanageChild( lightPointTextOk );
        XtUnmanageChild( lightAttenuation1 );
        XtUnmanageChild( lightAttenuation1Scroll );
        XtUnmanageChild( lightAttenuation1Label );
        XtUnmanageChild( lightAttenuation2 );
        XtUnmanageChild( lightAttenuation2Scroll );
        XtUnmanageChild( lightAttenuation2Label );
        XtUnmanageChild( lightConcentration );
        XtUnmanageChild( lightConcentrationScroll );
        XtUnmanageChild( lightConcentrationLabel );
        XtUnmanageChild( lightSpreadAngle );
        XtUnmanageChild( lightSpreadAngleScroll );
        XtUnmanageChild( lightSpreadAngleLabel );
        break;
    case 1:
        cur_light = DIRECTIONAL;
        XtVaSetValues( XtParent(XtParent(widget)), XtNlabel, "Vector", NULL );
        XtManageChild( lightDirection );
        XtManageChild( lightDirectionText );
        XtManageChild( lightDirectionTextOk );

        XtUnmanageChild( lightPoint );
        XtUnmanageChild( lightPointText );
        XtUnmanageChild( lightPointTextOk );
        XtUnmanageChild( lightAttenuation1 );
        XtUnmanageChild( lightAttenuation1Scroll );
        XtUnmanageChild( lightAttenuation1Label );
        XtUnmanageChild( lightAttenuation2 );
        XtUnmanageChild( lightAttenuation2Scroll );
        XtUnmanageChild( lightAttenuation2Label );
        XtUnmanageChild( lightConcentration );
        XtUnmanageChild( lightConcentrationScroll );
        XtUnmanageChild( lightConcentrationLabel );
        XtUnmanageChild( lightSpreadAngle );
        XtUnmanageChild( lightSpreadAngleScroll );
        XtUnmanageChild( lightSpreadAngleLabel );
        break;
    case 2:
        cur_light = POSITIONAL;
        XtVaSetValues( XtParent(XtParent(widget)), XtNlabel, "Point", NULL );
        XtManageChild( lightPoint );
        XtManageChild( lightPointText );
        XtManageChild( lightPointTextOk );
        XtManageChild( lightAttenuation1 );
        XtManageChild( lightAttenuation1Scroll );
        XtManageChild( lightAttenuation1Label );
        XtManageChild( lightAttenuation2 );
        XtManageChild( lightAttenuation2Scroll );
        XtManageChild( lightAttenuation2Label );

        XtUnmanageChild( lightDirection );
        XtUnmanageChild( lightDirectionText );
        XtUnmanageChild( lightDirectionTextOk );
        XtUnmanageChild( lightConcentration );
        XtUnmanageChild( lightConcentrationScroll );
        XtUnmanageChild( lightConcentrationLabel );
        XtUnmanageChild( lightSpreadAngle );
        XtUnmanageChild( lightSpreadAngleScroll );
        XtUnmanageChild( lightSpreadAngleLabel );
        break;
    case 3:
        cur_light = SPOT;
        XtVaSetValues( XtParent(XtParent(widget)), XtNlabel, "Spot", NULL );
        XtManageChild( lightDirection );
        XtManageChild( lightDirectionText );
        XtManageChild( lightDirectionTextOk );
        XtManageChild( lightPoint );
        XtManageChild( lightPointText );
        XtManageChild( lightPointTextOk );
        XtManageChild( lightAttenuation1 );
        XtManageChild( lightAttenuation1Scroll );
        XtManageChild( lightAttenuation1Label );
        XtManageChild( lightAttenuation2 );
        XtManageChild( lightAttenuation2Scroll );
        XtManageChild( lightAttenuation2Label );
        XtManageChild( lightConcentration );
        XtManageChild( lightConcentrationScroll );
        XtManageChild( lightConcentrationLabel );
        XtManageChild( lightSpreadAngle );
        XtManageChild( lightSpreadAngleScroll );
        XtManageChild( lightSpreadAngleLabel );
        break;
    }

    switch( lights[cur_light].color.type ) {
    case PEXColorTypeRGB:
        sprintf( string, "R: %.2f G: %.2f B: %.2f",
            lights[cur_light].color.value.rgb.red,
            lights[cur_light].color.value.rgb.green,
            lights[cur_light].color.value.rgb.blue );
        break;
    case PEXColorTypeHSV:
        sprintf( string, "H: %.2f S: %.2f V: %.2f",
            lights[cur_light].color.value.hsv.hue,
            lights[cur_light].color.value.hsv.saturation,
            lights[cur_light].color.value.hsv.value );
        break;
    case PEXColorTypeHLS:
        sprintf( string, "H: %.2f L: %.2f S: %.2f",
            lights[cur_light].color.value.hls.hue,
            lights[cur_light].color.value.hls.lightness,
            lights[cur_light].color.value.hls.saturation );
        break;
    case PEXColorTypeCIE:
        sprintf( string, "X: %.2f Y: %.2f Z: %.2f",
            lights[cur_light].color.value.cie.x,
            lights[cur_light].color.value.cie.y,
            lights[cur_light].color.value.cie.z );
        break;
    }
    XtVaSetValues( lightColorLabel, XtNlabel, string, NULL );
    free( string );

    if( cur_light == DIRECTIONAL ) {
        extern TextPtr text;
        string = (String) malloc( 30 );

        sprintf( string, "x: %.2f y: %.2f z: %.2f",
            lights[cur_light].direction.x,
            lights[cur_light].direction.y,
            lights[cur_light].direction.z );
        text->direction = XtNewString( string );
        XtVaSetValues( lightDirectionText,
            XtNstring, text->direction,
            XtNinsertPosition, strlen(text->direction),
            NULL );
        free( string );
    }

    if( cur_light == POSITIONAL ) {
        float topOfThumb;
        XtArgVal *l_topOfThumb;
        extern TextPtr text;
        string = (String) malloc( 30 );
    
        sprintf( string, "x: %.2f y: %.2f z: %.2f", lights[cur_light].point.x,
            lights[cur_light].point.y, lights[cur_light].point.z );
        text->point = XtNewString( string );
        XtVaSetValues( lightPointText,
            XtNstring, text->point,
            XtNinsertPosition, strlen(text->point),
            NULL );
    
        sprintf( string, "%.2f", lights[cur_light].attenuation1 );
        XtVaSetValues( lightAttenuation1Label, XtNlabel, string, NULL );
        sprintf( string, "%.2f", lights[cur_light].attenuation2 );
        XtVaSetValues( lightAttenuation2Label, XtNlabel, string, NULL );
    
        topOfThumb = lights[cur_light].attenuation1;
        l_topOfThumb = (XtArgVal *) &topOfThumb;
        XtVaSetValues( lightAttenuation1Scroll,
            XtNtopOfThumb, *l_topOfThumb, NULL );
        topOfThumb = lights[cur_light].attenuation2;
        l_topOfThumb = (XtArgVal *) &topOfThumb;
        XtVaSetValues( lightAttenuation2Scroll,
            XtNtopOfThumb, *l_topOfThumb, NULL );
        free( string );
    }

    if( cur_light == SPOT ) {
        float topOfThumb;
        XtArgVal *l_topOfThumb;
        extern TextPtr text;
        string = (String) malloc( 30 );

        sprintf( string, "x: %.2f y: %.2f z: %.2f",
            lights[cur_light].direction.x,
            lights[cur_light].direction.y,
            lights[cur_light].direction.z );
        text->direction = XtNewString( string );
        XtVaSetValues( lightDirectionText,
            XtNstring, text->direction,
            XtNinsertPosition, strlen(text->direction),
            NULL );

        sprintf( string, "x: %.2f y: %.2f z: %.2f", lights[cur_light].point.x,
            lights[cur_light].point.y, lights[cur_light].point.z );
        text->point = XtNewString( string );
        XtVaSetValues( lightPointText,
            XtNstring, text->point, 
            XtNinsertPosition, strlen(text->point),
            NULL );
    
        sprintf( string, "%.2f", lights[cur_light].attenuation1 );
        XtVaSetValues( lightAttenuation1Label, XtNlabel, string, NULL );
        sprintf( string, "%.2f", lights[cur_light].attenuation2 );
        XtVaSetValues( lightAttenuation2Label, XtNlabel, string, NULL );
    
        topOfThumb = lights[cur_light].attenuation1;
        l_topOfThumb = (XtArgVal *) &topOfThumb;
        XtVaSetValues( lightAttenuation1Scroll,
            XtNtopOfThumb, *l_topOfThumb, NULL );
        topOfThumb = lights[cur_light].attenuation2;
        l_topOfThumb = (XtArgVal *) &topOfThumb;
        XtVaSetValues( lightAttenuation2Scroll,
            XtNtopOfThumb, *l_topOfThumb, NULL );
        free( string );
    }
}


void editColor( widget, call_data, client_data )
Widget    widget;
XtPointer call_data, client_data;
{
    extern XtAppContext app;
    extern char *getenv();
    PEXColorSpecifier *colorSpec;
    String string = (String) malloc(40);

    if( getenv( "COLOR_EDITOR" ) ) {
        extern void get_reply();
        char *program, *args;

        XtAppAddInput( app, fileno(stdin), (XtPointer)XtInputReadMask, get_reply, NULL );

        switch ( cur_light ) {
            case AMBIENT:
                colorSpec = &lights[AMBIENT].color;
                break;
            case DIRECTIONAL:
                colorSpec = &lights[DIRECTIONAL].color;
                break;
            case POSITIONAL:
                colorSpec = &lights[POSITIONAL].color;
                break;
            case SPOT:
                colorSpec = &lights[SPOT].color;
                break;
        }

        program = (char *) malloc( strlen(getenv( "COLOR_EDITOR" )) + 25 );
        strcpy( program, getenv( "COLOR_EDITOR" ) );
        args = (char *) malloc(25);
        sprintf( args, " RGB %.2f %.2f %.2f", colorSpec->value.rgb.red, colorSpec->value.rgb.green, colorSpec->value.rgb.blue );
        strcat( program, args );
        xs_talkto( program );
    }
    else {
        fprintf( stderr, "Environment variable COLOR_EDITOR, undefined\n\n" );
        XtDestroyApplicationContext( app );
        exit( 0 );
    }
}


void get_reply(client_data, fid, id)
XtPointer    client_data;
int         *fid;
XtInputId   *id;
{
    char       buf[BUFSIZ];
    int        nbytes, i;

    /* Get all pending input */
    nbytes = read(*fid, buf, BUFSIZ);

    /*
     * Null terminate the string at the first newline,
     * or at the end of the bytes read.
     */
    if (nbytes && buf[0] != '\n') {
        for(i=0;i<nbytes;i++)
            if(buf[i] == '\n')
                buf[i] = '\0';
        buf[nbytes] = '\0';
        set_light_color( buf );
    }
}
